#include "stdafx.h"
#include "AStar2DGrid.h"					// CAStar2DGrid
#include <assert.h>								// assert()



CAStar2DGrid::CAStar2DGrid( const uint32 dwWidth, const uint32 dwHeight )
	:m_GridGraph(dwWidth,dwHeight),m_AStar(m_GridGraph)
{
}


CAStar2DGrid::CAStar2DGridGraph::CAStar2DGridGraph( const uint32 dwWidth, const uint32 dwHeight )
	:m_dwWidth(dwWidth), m_dwHeight(dwHeight)
{
	m_Nodes.resize(m_dwWidth*m_dwHeight);
}

void CAStar2DGrid::CAStar2DGridGraph::SetGoal( const SArrayPos v ) 
{
	m_Goal = v;
}

void CAStar2DGrid::CAStar2DGridGraph::SetObstacle( const SArrayPos v )
{	
	assert((uint32)v.m_X<m_dwWidth);
	assert((uint32)v.m_Y<m_dwHeight);

	CArrayNode &r = m_Nodes[v.m_X+v.m_Y*m_dwWidth];

	assert(r.m_ParentAndType<=0);		// should only be applied when map is clean

	r.m_ParentAndType=-1;
}


void CAStar2DGrid::CAStar2DGridGraph::ClearObstacle( const SArrayPos v )
{
	assert((uint32)v.m_X<m_dwWidth);
	assert((uint32)v.m_Y<m_dwHeight);

	CArrayNode &r = m_Nodes[v.m_X+v.m_Y*m_dwWidth];

	assert(r.m_ParentAndType<=0);		// should only be applied when map is clean

	r.m_ParentAndType=0;
}

bool CAStar2DGrid::CAStar2DGridGraph::GetObstacle( const SArrayPos v ) const
{
	assert((uint32)v.m_X<m_dwWidth);
	assert((uint32)v.m_Y<m_dwHeight);

	const CArrayNode &r = m_Nodes[v.m_X+v.m_Y*m_dwWidth];

	return r.m_ParentAndType == -1;
}

void CAStar2DGrid::CAStar2DGridGraph::OnPushNode( const SArrayPos v )
{
	// compute heuristic

	CArrayNode &rNode = m_Nodes[v.m_X+v.m_Y*m_dwWidth];

	if(rNode.m_ParentAndType==0)					// start
		rNode.m_ParentAndType=5;

	short X = m_Goal.m_X-v.m_X;
	short Y = m_Goal.m_Y-v.m_Y;

	short L = max(abs(X),abs(Y))*100;		// manhatten distance

	rNode.m_CostPlusHeuristic = rNode.m_Cost + L;
}

bool CAStar2DGrid::CAStar2DGridGraph::IsWalkableAndNotKnown( const SArrayPos v ) const
{
	if((uint32)v.m_X>=m_dwWidth || (uint32)v.m_Y>=m_dwHeight)
		return false;

	return m_Nodes[v.m_X+v.m_Y*m_dwWidth].m_ParentAndType == 0;
}


CAStar2DGrid::SArrayPos CAStar2DGrid::CAStar2DGridGraph::GetPrevious( const CAStar2DGrid::SArrayPos v ) const
{
	char ParentId = m_Nodes[v.m_X+v.m_Y*m_dwWidth].m_ParentAndType;

	switch(ParentId)
	{
		case 2: return SArrayPos(v.m_X  ,v.m_Y-1);
		case 4: return SArrayPos(v.m_X-1,v.m_Y  );
		case 6: return SArrayPos(v.m_X+1,v.m_Y  );
		case 8: return SArrayPos(v.m_X  ,v.m_Y+1);
		case 5:	return SArrayPos();										// start reached
		default: assert(0);
	}
	

	return SArrayPos();
}


CAStar2DGrid::SArrayPos CAStar2DGrid::Solve()
{
	while(CAStar2DGrid::SArrayPos v = m_AStar.Process())					// find a good node to process
		if(IsGoal(v))
			return v;

	return CAStar2DGrid::SArrayPos();
}



bool CAStar2DGrid::CAStar2DGridGraph::WasNodeTested( const SArrayPos v ) const
{
	if((uint32)v.m_X>=m_dwWidth || (uint32)v.m_Y>=m_dwHeight)
		return false;

	return m_Nodes[v.m_X+v.m_Y*m_dwWidth].m_ParentAndType >1;
}



bool CAStar2DGrid::CAStar2DGridGraph::ComparePaths( const SArrayPos l, const SArrayPos r ) const
{
	const CArrayNode &rL = m_Nodes[l.m_X+l.m_Y*m_dwWidth];
	const CArrayNode &rR = m_Nodes[r.m_X+r.m_Y*m_dwWidth];

	return rL.m_CostPlusHeuristic > rR.m_CostPlusHeuristic;
}


bool CAStar2DGrid::CAStar2DGridGraph::IsGoal( const SArrayPos v ) const
{
	return v==m_Goal;
}


